"""Infoseek DA"""
__version__='$Revision: 1.9 $'[11:-2]

import sys, os, time, httplib, rotor
import marshal, md5, base64
from Globals import HTMLFile, MessageDialog
from AccessControl.Role import RoleManager
from Shared.DC.ZRDB.Results import Results
from App.Extensions import getBrain
from OFS.SimpleItem import Item
from Globals import Persistent
from Acquisition import Implicit
from ImageFile import ImageFile
from cStringIO import StringIO
from urllib import quote, unquote
from zlib import compress, decompress
from Shared.DC.ZRDB.Aqueduct import decodestring, parse
from string import strip, find, atoi, split, join, upper, lower

from DocumentTemplate import DT_Util

import string, mimetools, re

class HTTP(httplib.HTTP):
    # A revised version of the HTTP class that can do basic
    # HTTP 1.1 connections, and also compensates for a bug
    # that occurs on some platforms in 1.5 and 1.5.1 with
    # socket.makefile().read()

    read_bug=sys.version[:5] in ('1.5 (','1.5.1')

    def putrequest(self, request, selector, ver='1.0'):
        selector=selector or '/'
        str = '%s %s HTTP/%s\r\n' % (request, selector, ver)
        self.send(str)

    def getreply(self):
        file=self.sock.makefile('rb')
        data=string.join(file.readlines(), '')
        file.close()
        self.file=StringIO(data)
        line = self.file.readline()
        try:
            [ver, code, msg] = string.split(line, None, 2)
        except ValueError:
          try:
              [ver, code] = string.split(line, None, 1)
              msg = ""
          except ValueError:
              return -1, line, None
        if ver[:5] != 'HTTP/':
            return -1, line, None
        code=string.atoi(code)
        msg =string.strip(msg)
        headers =mimetools.Message(self.file, 0)
        return code, msg, headers


class UltraseekDA(Item, Persistent, Implicit, RoleManager):
    """Ultraseek DA"""

    meta_type='Ultraseek DA'
    icon     ='misc_/UltraseekDA/UltraseekDA_icon'

    _col=None
    max_rows_=1000
    cache_time_=0
    max_cache_=100
    rotor=None
    key=''
    class_file_=''
    class_name_=''

    manage_options=(
    {'label':'Status',   'action':'manage_main'},
    {'label':'Properties', 'action':'manage_properties'},
    {'label':'Security',   'action':'manage_access'},
    )

    __ac_permissions__=(
    ('View management screens',
     ('manage_main', 'manage_tabs', 'manage_properties')),
    ('Change Ultraseek DAs', ('manage_edit','manage_properties',
                             'generate_default_docs')),
    ('Change permissions', ('manage_access')),
    ('Use Ultraseek DAs', ('__call__',''), ('Anonymous', 'Manager')),
    )

    manage_main=HTMLFile('main', globals())
    manage_properties=HTMLFile('properties', globals())

    def __init__(self, id, host, key, title):
        self.id=id
        self.manage_edit(host, key, title)

    def __setstate__(self, state):
	Persistent.__setstate__(self, state)
	if self.host:
	    try:    self.get_config()
	    except: pass

    def manage_edit(self, host, key, title='', REQUEST=None):
	"""Change properties"""
	self.title=title
        self.host, self.root, v=parse_connection(host)
	self.key=key
	self.rotor=Rotor(key)
        try:    self.get_config(r=1)
        except: self.connect_error()
	if REQUEST is not None:
	    return MessageDialog(title='Properties updated',
                                 message='Properties updated.',
                                 action='%s/manage_main' % REQUEST['URL1']
                                 )

    _special_var_names=('config', )
    
    def _argdata(self, dicts):
        # Create argdata from one or more dicts
        svar=self._special_var_names
        data={}
        for dict in dicts:
            for k, v in dict.items():
                # Filter out special vars
                if upper(k) != k and k not in svar:
                    data[k]=v
        return querify_dict(data)

    _v_config_data=None

    def get_config(self, r=0):
        if self._v_config_data is None or r:
            urlpath='/%s/config.html' % self.root
            data, flag=self._fetch(urlpath)
            data=marshal.loads(data)
            self._v_config_data=config_data(data)
        return self._v_config_data

    def get_query(self, argdata):
        urlpath='/%s/query.html' % self.root
        data, flag=self._fetch(urlpath, argdata)
        data=marshal.loads(data)
        return data

    def __call__(self, REQUEST=None, __ick__=None, src__=0, **kw):
	"""Call the remote Ultraseek server.

	The arguments to the method should be passed via keyword
	arguments, and/or the REQUEST object. Form variables from
        the REQUEST will be combined with passed keyword arguments
        to form the query string used to contact the Infoseek
        server.

	The returned value is a sequence of record objects.
	"""

        if not self.host:
            self.connect_error('invalid server')
        
        argdata=[]
	if REQUEST is not None:
            argdata.append(REQUEST.other)
        if kw: argdata.append(kw)
        argdata=self._argdata(argdata)

        # TODO: add caching code here later

        result=self.get_query(argdata)

        __traceback_info__=(argdata, result)

        if hasattr(self, '_v_brain'):
            pb=self._v_brain
        else:
            pb=self._v_brain=getBrain(self.class_file_, self.class_name_)

        result=Results(result, pb, self)
	columns=result._searchable_result_columns()
	if columns != self._col: self._col=columns
	return result

    index_html=__call__

    def _fetch(self, url, argdata=''):
        # Ultraseek doesnt allow us to use exact pnp because
        # it doesn't like POSTs with funny content-types...

        if not self.host:
            self.connect_error('invalid server')

        if argdata:
            url='%s?%s' % (url, argdata)

	h=HTTP(self.host)
	try:    h.putrequest('GET', url)
	except: self.config_error()
        h.putheader('Host', self.host)
        h.putheader('Accept', '*/*')
        h.putheader('User-Agent', 'Ultraseek DA %s' % __version__)
	h.endheaders()
	errcode,errmsg,headers=h.getreply()
        file=h.getfile()
	data=strip(file.read())
	file.close()
        ct=lower(headers['content-type'])
        flag=0
        if ct in ('application/x-zope-network',
                  'application/x-principia-network',
                  'text/x-pydb'):
            flag=1
            data=base64.decodestring(data)
            data=self.rotor.decrypt(data)
            digest,data=data[:16],data[16:]
            if md5.new(data).digest() != digest:
                raise 'Bad Request', 'Corrupted Data'
        if ct=='application/x-python-marshal':
            data=base64.decodestring(data)
        
	if errcode != 200:
	    __traceback_info__=data, headers.__dict__, argdata, errcode
            self.connect_error(data)
        return data, flag

    def connect_error(self, msg=None):
        if msg is None:
            msg='%s: %s' % (sys.exc_type, sys.exc_value)
        raise 'Connection error', (
            "<em>An error occurred connecting to the server.</em>\n"
	    "<!--\n%s\n-->"
	    % msg)


    def generate_default_docs(self, REQUEST=None):
	"""Generate default interface documents in the current folder."""
        fsdir=os.path.join(SOFTWARE_HOME, 'Products/UltraseekDA/default_docs')
        obdir=self.aq_parent
        cpy_docs(fsdir, obdir, self.id)
	if REQUEST is not None:
	    return MessageDialog(title='Documents created',
                                 message='Documents created.',
                                 action='%s/manage_main' % REQUEST['URL1']
                                 )

    def termpat(self, dterms): 
        """ Emulate Ultraseek's collectn.termpat (or, what it seems to do)
        """
        terms=map(lambda x: x[0], dterms) 
        pat=r'\b(%s)\b' % (string.join(terms, '|'))
        return re.compile(pat, re.I)

    def html_quote(self, *args):
        """ Provide access from in DA documents to html_quote.  Ugly. """ 
        return apply(DT_Util.html_quote, args)
    
    
class Rotor:
    def __init__(self, key, numrotors=6):
	self.a=key, numrotors
	r=rotor.newrotor(key, numrotors)
	self.encrypt=r.encrypt
	self.decrypt=r.decrypt
	self.decryptmore=r.decryptmore

    def __getinitargs__(self): return self.a
    def __getstate__(self,v={}): return v


class collection:
    # A surrogate for an Ultraseek collection object
    def __init__(self, id, name):
        self.id=id
        self.pretty_name=name


class config_data:
    def __init__(self, data):
        cols={}
        for id, name in data['collections']:
            cols[id]=collection(id, name)
        data['collections']=cols
        self.data=data

    def __getattr__(self, n):
        data=self.__dict__['data']
        if data.has_key(n):
            return data[n]
        raise AttributeError, n


manage_addUltraseekDAForm=HTMLFile('add', globals())
def manage_addUltraseekDA(self, id, host, key, title='', REQUEST=None):
    """Add object"""
    ob=UltraseekDA(id, host, key, title)
    self._setObject(id, ob)
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)


def parse_connection(value):
    value=strip(value)
    while value[-1:]=='/': value=value[:-1]
    url=value[7:]
    l=find(url,'/')
    if lower(value)[:7]=='http://' and l > 0:
	server=url[:l]
	base=url[l+1:]
    else:
	raise 'BadRequest', (
	    '<strong>The value entered:<br>%s<p>'
	    'is not an absolute URL' % value)
    return  server, base, value


def cpy_docs(fsdir, obdir, da_name, path_join=os.path.join):
    """Generate interface documents"""
    for name in os.listdir(fsdir):
        pathname=path_join(fsdir, name)

        if pathname[-1]=='.':
            continue
        
        elif os.path.isfile(pathname):
            if name[-5:]=='.dtml':
                file=open(pathname, 'r')
                data=file.read()
                file.close()
                # Insert correct DA name :(
                data=join(split(data, 'DA_NAME'), da_name)
                obdir.manage_addDocument(name[:-5]+'_html', '', data)

            if name[-4:] in ('.gif', '.jpg'):
                file=open(pathname, 'rb')
                data=file.read()
                file.close()
                obdir.manage_addImage(name, data)

        elif os.path.isdir(pathname):
            obdir.manage_addFolder(name)
            cpy_docs(path_join(fsdir, name), getattr(obdir, name), da_name)

        else:
            __traceback_info__=(fsdir, obdir, name)
            raise ValueError, name





# Python -> HTTP marshalling - note that this is different
# than that used by bci, as non-bobo platforms dont know
# how to use the type info included in the bci version!

def marshal_str(n, v):
    return '%s=%s' % (n, quote(v))

def marshal_int(n, v):
    return '%s=%s' % (n, v)

marshal_float=marshal_int

def marshal_long(n, v):
    return ('%s=%s' % (n, v))[:-1]

def marshal_list(n, v, lt=type([]), tt=type(())):
    ret=[]
    for ob in v:
	tp=type(ob)
	if tp is lt or tp is tt:
            __traceback_info__=(n, v)
	    raise TypeError, 'Invalid recursion in data to be marshaled.'
        ret.append(marshal_object(n, ob))
    return join(ret, '&')

marshal_tuple=marshal_list

type2marshal={
    type('s'): marshal_str,
    type(1):   marshal_int,
    type(1.0): marshal_float,
    type(1L):  marshal_long,
    type([]):  marshal_list,
    type(()):  marshal_tuple,
    }

def marshal_object(k, v, has_key=type2marshal.has_key):
    tp=type(v)
    return has_key(tp) and type2marshal[tp](k, v) or None

def querify_dict(dict):
    keys=dict.keys()
    keys.sort()
    query=[]
    for key in keys:
        m=marshal_object(key, dict[key])
        if m: query.append(m)
    return query and join(query, '&') or ''
